#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import json

from aiuna.step.file import File
from tatu.okast import OkaSt
from aiuna.content.creation import read_arff

#with open("token.txt", "r") as f:
    #token = json.load(f)["token"]

#storage = OkaSt(token=token)

#print("Reading file...")
#data = File("iris.arff")

#print("Storing...")
#storage.store(data)  # TODO: it is always sending the file, even when not needed

#print("Fetching...")
#d = storage.fetch(data.id)
#print(d)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import json
import os
import traceback
from glob import glob
from pathlib import Path

import _pickle as pickle

from aiuna.content.data import Data
from garoupa.uuid import UUID
from tatu.abs.storage import MissingEntryException, LockedEntryException, DuplicateEntryException, \
    UnlockedEntryException
from tatu.disk import save, load
from tatu.storageinterface import StorageInterface


class Pickle(StorageInterface):
    def _config_(self):
        return self._config

    def _hasdata_(self, id, include_empty):
        pass

    def _hasstream_(self, data):
        pass

    def _getdata_(self, id, include_empty):
        pass

    def _getstream_(self, data):
        pass

    def _hasstep_(self, id):
        pass

    def _getstep_(self, id):
        pass

    def _getfields_(self, id):
        pass

    def _hascontent_(self, ids):
        pass

    def _getcontent_(self, id):
        pass

    def _lock_(self, id):
        pass

    def _putdata_(self, id, step, inn, stream, parent, locked, ignoredup):
        pass

    def _putstream_(self, rows, ignoredup):
        pass

    def _putfields_(self, rows, ignoredup):
        pass

    def _putcontent_(self, id, value, ignoredup):
        pass

    def _putstep_(self, id, name, path, config, dump, ignoredup):
        pass

    def _deldata_(self, id):
        pass

    def _fetch_children_(self, data: Data):
        raise Exception("Storage Pickle cannot retrieve children!")

    def __init__(self, threaded=True, db="tatu-sqlite", compress=True, close_when_idle=False):
        self._config = locals().copy()
        del self._config["self"]
        del self._config["__class__"]
        self.db = db
        self.compress = compress
        if not Path(db).exists():
            os.mkdir(db)
        self._uuid = UUID(json.dumps([db, compress], sort_keys=True, ensure_ascii=False).encode())
        super().__init__(threaded, timeout=1, close_when_idle=close_when_idle)

    def _uuid_(self):
        return self._uuid

    def _delete_(self, data: Data, check_missing=True):
        locked = self._filename("", data)
        if check_missing and not Path(locked).exists():
            raise MissingEntryException("Does not exist:", data.id)
        os.remove(locked)

    def _fetch_(self, data: Data, lock=False):
        # TODO: deal with fields and missing fields?
        filename = self._filename("*", data)

        # Not started yet?
        if not Path(filename).exists():
            print('W: Not started.', filename)
            if lock:
                print("W: Locking...", filename)
                Path(filename).touch()
            return None

        # Locked?
        if Path(filename).stat().st_size == 0:
            print("W: Previously locked by other process.", filename)
            raise LockedEntryException(filename)

        transformed_data = self._load(filename)

        # # Failed?
        # if transformed_data.failure is not None:
        #     raise FailedEntryException(transformed_data.failure)

        return transformed_data

    def _store_(self, data, check_dup=True):
        """The dataset name of data_out will be the filename prefix for
        convenience."""

        # TODO: reput name on Data?
        filename = self._filename("", data)
        # filename = self._filename(data.name, data, training_data_uuid)

        # sleep(0.020)  # Latency simulator.

        # Already exists?
        if check_dup and Path(filename).exists():
            raise DuplicateEntryException("Already exists:", filename)

        locked = self._filename("", data)
        if Path(locked).exists():
            os.remove(locked)

        self._dump(data, filename)

    def _filename(self, prefix, data):
        zip = "compressed" if self.compress else ""
        # Not very efficient.  TODO: memoize extraction of fields from JSON?
        # uuids = [json.loads(tr)['uuid'][:6] for tr in data.history]
        # rest = f"-".join(uuids) + f".{zip}.dump"
        rest = f"{data if isinstance(data, str) else data.id}.{zip}.dump"
        if prefix == "*":
            query = self.db + "/*" + rest
            lst = glob(query)
            if len(lst) > 1:
                raise Exception("Multiple files found:", query, lst)
            if len(lst) == 1:
                return lst[0]
            else:
                return self.db + "/" + rest
        else:
            return self.db + "/" + prefix + rest

    def _load(self, filename):
        """
        Retrieve a Data object from disk.
        :param filename: file dataset
        :return: Data
        """
        try:
            if self.compress:
                data = load(filename)
            else:
                f = open(filename, "rb")
                res = pickle.load(f)
                f.close()
                data = res
            return data
        except Exception as e:
            traceback.print_exc()
            print("Problems loading", filename)
            exit(0)

    def _dump(self, data, filename):
        """
        Dump a Data object to disk.
        :param data: Data
        :param filename: file dataset
        :return: None
        """
        print("W: Storing...", filename)
        data = data.picklable
        if self.compress:
            save(filename, data)
        else:
            f = open(filename, "wb")
            pickle.dump(data, f)
            f.close()

    def _unlock_(self, data):
        filename = self._filename("*", data)
        if not Path(filename).exists():
            raise UnlockedEntryException("Cannot unlock something that is not " "locked!", filename)
        print("W: Unlocking...", filename)
        os.remove(filename)

    def _open_(self):
        pass

    def _close_(self):
        pass

    def fetch_field(self, _id):
        raise NotImplementedError

    # def list_by_name(self, substring, only_original=True):
    #     datas = []
    #     path = self.db + f"/*{substring}*-*.dump"
    #     for file in sorted(glob(path), key=os.path.getmtime):
    #     data = self._load(file)
    #     if only_original and len(data.history) == 1:
    #         datas.append(data)  # Data is now fully lazy, so nothing will be retrieved.
    #     return datas


# def sqlite_Test():
#     from tatu.sql.sqlite import SQLite
#     from aiuna.step.file import File
#     data = File("iris.arff").data
#     SQLite().delete(data, check_missing=False)
#     SQLite().store(data)
#     print(SQLite().visual_history(File("iris.arff").data.id))

# from tatu.sql.sqlite import SQLite
# PickleServer().store(File("iris.arff").data)
# print(PickleServer().visual_history(File("iris.arff").data.id))
# exit()

# from tatu.sql.mysql import MySQL
# MySQL(db="tatu:xxxxx@localhost/tatu").store(File("iris.arff").data)
# print(MySQL(db="tatu:xxxxxx@localhost/tatu").visual_history(File("iris.arff").data.id))
# exit()

# sqlite_Test()



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#
#      tatu is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      tatu is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
from sqlalchemy.engine import Engine

from tatu.abs.storage import Storage


class Tatu(Storage):
    storage = None

    def _config_(self):
        return self._config

    def __init__(self, url="sqlite://tatu-sqlite", threaded=True, alias=None, close_when_idle=False):
        self._config = locals().copy()
        del self._config["self"]
        if "__class__" in self._config:
            del self._config["__class__"]
        # print("STORAGE:", url)
        if isinstance(url, Engine):
            from tatu.sql.sqla import SQLA
            self.storage = SQLA(engine=url)
        else:
            if "://" not in url:
                raise Exception("Missing '://' in url:", url)
            backend, db = url.split("://")
            if backend == "sqlite":
                from tatu.sql.sqlite import SQLite
                self.storage = SQLite(db, threaded, close_when_idle)
            elif backend == "mysql":
                from tatu.sql.mysql import MySQL
                self.storage = MySQL(db, threaded, close_when_idle)
            elif backend == "oka":
                from tatu.okast import OkaSt
                token, url = db.split("@")
                self.storage = OkaSt(token, alias, threaded, "http://" + url,
                                     close_when_idle)  # TODO Accept user/login in OkaSt?
            else:
                raise Exception("Unknown DBMS backend:", url)

    def _uuid_(self):
        return self.storage.uuid

    # TODO make all args/kwargs explicity for better docs/IDE integration
    def fetch(self, data, lock=False, lazy=True):
        return self.storage.fetch(data, lock, lazy)

    def store(self, data, unlock=False, ignoredup=False, lazy=False):
        return self.storage.store(data, unlock, ignoredup, lazy)

    def fetchhistory(self, id):
        return self.storage.fetchhistory(id)

    def fetchstep(self, id):
        return self.storage.fetchstep(id)

    def hasdata(self, id, include_empty=False):
        return self.storage.hasdata(id, include_empty)

    def hasstream(self, data):
        return self.storage.hasstream(data)

    def getdata(self, id, include_empty=True):
        return self.storage.getdata(id, include_empty)

    def getstream(self, data):
        return self.storage.getstream(data)

    def hasstep(self, id):
        return self.storage.hasstep(id)

    def getstep(self, id):
        return self.storage.getstep(id)

    def getfields(self, id):
        return self.storage.getfields(id)

    def getcontent(self, id):
        return self.storage.getcontent(id)

    def hascontent(self, ids):
        return self.storage.hascontent(ids)

    def removedata(self, data, check_existence=True, recursive=True):
        return self.storage.removedata(data, check_existence, recursive)

    def lock(self, id, check_existence=True):
        return self.storage.lock(id, check_existence)

    def deldata(self, id, check_success=True):
        return self.storage.deldata(id, check_success)

    def unlock(self, id, check_success=True):
        return self.storage.unlock(id, check_success)

    def putdata(self, id, step, inn, stream, parent, locked, ignoredup=False):
        return self.storage.putdata(id, step, inn, stream, parent, locked, ignoredup)

    def putstream(self, rows, ignoredup=False):
        return self.storage.putstream(rows, ignoredup)

    def putcontent(self, id, value, ignoredup=False):
        return self.storage.putcontent(id, value, ignoredup)

    def putfields(self, rows, ignoredup=False):
        return self.storage.putfields(rows, ignoredup)

    def storestep(self, step, dump=None, ignoredup=False):
        return self.storage.storestep(step, dump, ignoredup)

    def putstep(self, id, name, path, config, dump=None, ignoredup=False):
        return self.storage.putstep(id, name, path, config, dump, ignoredup)

    def open(self):
        return self.storage.open()

    def close(self):
        return self.storage.close()



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
import json
from abc import ABC, abstractmethod

from aiuna.compression import unpack, fpack
from aiuna.content.data import Data
from aiuna.content.root import Root
from aiuna.history import History
from garoupa.uuid import UUID
from akangatu.linalghelper import islazy
from tatu.abs.mixin.thread import asThread
from tatu.abs.storage import Storage, DuplicateEntryException, LockedEntryException
from akangatu.transf.step import Step


class StorageInterface(asThread, Storage, ABC):
    # TODO (strict) fetch by uuid
    # TODO (strict) store
    def fetch(self, data, lock=False, lazy=True):
        # , recursive=True):   # TODO: pensar no include_empty=False se faz sentido
        """Fetch the data object fields on-demand.
         data: uuid string or a (probably still not fully evaluated) Data object."""
        data_id = data if isinstance(data, str) else data.id
        # lst = []
        # print("LOGGING:::  Fetching...", data_id)
        # while True:
        ret = self.getdata(data_id, include_empty=True)

        if ret is None or not ret["uuids"]:
            # REMINDER: check_existence false porque pode ser um data vazio
            # [e é para o Cache funcionar mesmo que ele tenha sido interrompido]
            if lock and not self.lock(data_id, check_existence=False):
                raise Exception("Could not lock data:", data_id)
            return

        fields = {} if isinstance(data, str) else data.field_funcs_m
        for field, fid in list(ret["uuids"].items()) + [("stream", None)]:
            if field == "inner":
                fields[field] = lambda: self.fetch(fid)
            elif field == "stream":
                fields[field] = lambda: self.fetchstream(data_id, lazy)
            elif field == "changed":
                fields[field] = unpack(self.getcontent(fid)) if isinstance(data, str) else data.changed
            elif field not in ["inner"] and (isinstance(data, str) or field in data.changed):
                if lazy:
                    fields[field] = (lambda fid_: lambda: unpack(self.getcontent(fid_)))(fid)
                else:
                    fields[field] = unpack(self.getcontent(fid))

            # Call each lambda by a friendly name.
            if lazy and field != "changed":
                fields[field].__name__ = "_" + fields[field].__name__ + "_from_storage_" + self.id

        if isinstance(data, str):
            if lazy:
                raise Exception("lazy ainda não retorna histórico para data dado por uuid-string")  # <-- TODO
            history = self.fetchhistory(data)
        else:
            history = data.history
        # print("LOGGING:::  > > > > > > > > > fetched?", data_id, ret)
        return Data(UUID(data_id), {k: UUID(v) for k, v in ret["uuids"].items()}, history, **fields)

    def store(self, data: Data, unlock=False, ignoredup=False, lazy=False):
        """Store all Data object fields as soon as one of them is evaluated.

        # The sequence of queries is planned to minimize traffic and CPU load,
        # otherwise it would suffice to just send 'insert or ignore' of dumps.

        Parameters
        ----------
        data
            Data object to store.
        ignore_dup
            Whether to send the query anyway, ignoring errors due to already existent registries.

        Returns
        -------
        List of inserted (or hoped to be inserted for threaded storages) Data ids (only meaningful for Data objects with inner)

        Exception
        ---------
        DuplicateEntryException
        :param ignoredup:
        :param data:
        :param check_dup:
        :param recursive:
        """
        if not ignoredup and self.hasdata(data.id):
            raise DuplicateEntryException(data.id)

        # Embed lazy storers inside the Data object.
        lst = []

        def func(held_data, name, field_funcs, puts):
            def lamb():
                for k, v in field_funcs.items():
                    if islazy(v):
                        v = v()  # Cannot call data[k] due to infinite loop.
                    held_data.field_funcs_m[
                        k] = v  # The old value may not be lazy, but the new one can be due to this very lazystore.
                    id = held_data.uuids[k].id
                    if id in puts:
                        if k != "inner":
                            # TODO/REMINDER: exceptionally two datasets can have some equal contents, like Xd;
                            #   so we send it again while the hash is not based on content
                            self.putcontent(id, fpack(held_data, k), ignoredup=True)
                rows = [(held_data.id, fname, fuuid.id) for fname, fuuid in held_data.uuids.items() if fname != "inner"]
                self.putfields(rows)
                return held_data.field_funcs_m[name]

            return lamb

        streams = {}
        while True:
            # Fields.
            cids = [u.id for u in data.uuids.values()]
            missing = [cid for cid in cids if cid not in self.hascontent(cids)]
            if lazy:
                # All fields will be evaluated at the time one of them is called.
                field_funcs_copy = data.field_funcs_m.copy()
                for field in data.field_funcs_m:
                    if field == "stream":
                        # data.field_funcs_m["stream"] = map(
                        #     lambda d: self.store(d, unlock=True, lazy=False),
                        #     data.field_funcs_m["stream"]
                        # )
                        raise Exception("A lazy storage cannot handle streams for now.")
                    data.field_funcs_m[field] = func(data, field, field_funcs_copy, missing)
                    data.field_funcs_m[field].__name__ = "_" + data.uuids[field].id + "_to_storage_" + self.id
            else:
                for k, v in data.items():
                    if k == "stream" and data.stream and len(list(data.stream)) > 0:
                        # Consume stream, to be stored after putdata().
                        streams[data.id] = list(data.stream)
                    else:
                        id = data.uuids[k].id
                        if id in missing and k != "inner" and k != "stream":
                            content = fpack(data, k)
                            # TODO/REMINDER: exceptionally two datasets can have some equal contents, like Xd;
                            #   so we send it again while the hash is not based on content
                            self.putcontent(id, content, ignoredup=True)

            lst.append(data)
            if not data.hasinner:
                break
            data = data.inner

        for i, d in reversed(list(enumerate(lst))):
            if i == 0 and unlock:
                self.unlock(d.id)

            # History.
            ancestor_duuid = Root.uuid
            for step in list(d.history):
                # print("LOGGING:::  ssssssssSSSSSSSSSSSS", step.id)
                if not self.hasstep(step.id):
                    self.storestep(step)

                parent_uuid = ancestor_duuid
                ancestor_duuid = ancestor_duuid * step.uuid
                # Here, locked=NULL means 'placeholder', which can be updated in the future if the same data happens to be truly stored.
                # We assume it is faster to do a single insertignore than select+insert, hence ignoredup=True here.
                if ancestor_duuid == d.uuid:
                    hasstream, inner = d.hasstream, d.inner if d.hasinner else None
                else:
                    hasstream, inner = False, None
                self.putdata(ancestor_duuid.id, step.id, inner and inner.id, hasstream, parent_uuid.id, None,
                             ignoredup=True)
                # TODO: adopt logging    print(datauuid, 3333333333333333333333333333333333333333)

            if lazy:
                # TODO: adopt logging    print(d.id, 7777777777777777777777777777777)
                pass
            else:
                if d.id in streams:
                    rows = []
                    for pos, streamed_data in enumerate(streams[d.id]):
                        self.store(streamed_data, ignoredup=True)
                        rows.append((d.id, str(pos), streamed_data.id))
                    if not rows:
                        raise Exception("Empty stream??")
                    self.putstream(rows)

                    # Return a new iterator in the place of the original stream.
                    d.field_funcs_m["stream"] = iter(streams[d.id])
                self.putfields([(d.id, fname, fuuid.id) for fname, fuuid in d.uuids.items()], ignoredup=True)
        return lst[0]

    def fetchhistory(self, id):
        # print("LOGGING:::  Fetching history...", id)
        steps = []
        while True:
            ret = self.getdata(id)
            steps.append(self.fetchstep(ret["step"]))
            id = ret["parent"]
            if id == UUID().id:
                break
        history = History(steps[-1])
        for step in reversed(steps[:-1]):
            history <<= step
        # print("LOGGING:::     ...history fetched!", id)
        return history

    def fetchstream(self, dataid, lazy=True):
        """Return a stream iterator or None."""
        # print("LOGGING:::  Fetching step...", id)
        rows = self.getstream(dataid)
        # print("LOGGING:::         ...fetched step?", id, bool(r), r)
        if rows is None:
            return None
        for r in rows:
            yield self.fetch(r["chunk"], lazy=lazy)

    def fetchstep(self, id):
        """Return a Step object or None."""
        # print("LOGGING:::  Fetching step...", id)
        r = self.getstep(id)
        # print("LOGGING:::         ...fetched step?", id, bool(r), r)
        if r is None:
            return None
        r["config"] = json.loads(r["config"])
        return Step.fromdict({"id": id, "desc": r})

    def hasdata(self, id, include_empty=False):
        """ include_empty: whether to assess the existence of fields, instead of just the data row"""
        return self.do(self._hasdata_, locals(), wait=True)

    def getdata(self, id, include_empty=True):
        """Return a info for a Data object."""
        # print("LOGGING:::  Getting data...", id)
        r = self.do(self._getdata_, locals(), wait=True)
        # print("LOGGING:::         ...got data?", id, bool(r))
        # TODO: adopt logging    print(r)
        return r

    def hasstream(self, data):
        """Verify existence of a stream for the given dataid."""
        return self.do(self._hasstream_, locals(), wait=True)

    def getstream(self, data):
        """Return rows with info for the stream of a given Data id."""
        # print("LOGGING:::  Getting data...", id)
        r = self.do(self._getstream_, locals(), wait=True)
        # print("LOGGING:::         ...got data?", id, bool(r))
        # TODO: adopt logging    print(r)
        return r

    def hasstep(self, id):
        return self.do(self._hasstep_, locals(), wait=True)

    def getstep(self, id):
        """Return info for a Step object."""
        # print("LOGGING:::  Getting step...", id)
        r = self.do(self._getstep_, locals(), wait=True)
        # print("LOGGING:::         ...got step?", id, bool(r))
        return r

    # REMINDER we check missing fields through hascontent()
    def getfields(self, id):
        """Return fields and content for a Data object."""
        # print("LOGGING:::  Getting fields...", id)
        r = self.do(self._getfields_, locals(), wait=True)
        # print("LOGGING:::         ...got fields?", id, bool(r))
        return r

    def getcontent(self, id):
        """Return content."""
        # print("LOGGING:::  Getting content...", id)
        r = self.do(self._getcontent_, locals(), wait=True)
        # print("LOGGING:::         ...got content?", id, bool(r))
        return r

    def hascontent(self, ids):
        return self.do(self._hascontent_, {"ids": ids}, wait=True)

    def removedata(self, data: Data, check_existence=True, recursive=True):
        """Remove Data object, but keeps contents of its fields (even if not used by anyone else).

        Returns list of deleted Data object uuids
        """
        # print("LOGGING:::  Deleting...", data.id)
        ids = []
        while True:
            id = data.id
            self.deldata(id)
            ids.append(id)
            if not recursive or not data.hasinner:
                break
            data = data.inner
        # print("LOGGING:::           ...deleted!", ids)
        return ids

    def lock(self, id, check_existence=True):
        """Return whether it succeeded."""
        # print("LOGGING:::  Locking...", id)
        if check_existence and self.hasdata(id):
            raise Exception("Cannot lock, data already exists:", id)
        r = self.do(self._lock_, {"id": id}, wait=True)
        # print("LOGGING:::      ...locked?", id, bool(r))
        return r

    def deldata(self, id, check_success=True):
        """Return whether it succeeded."""
        # print("LOGGING:::  Deleting data...", id)
        r = self.do(self._deldata_, {"id": id}, wait=True)
        if check_success and not r:
            raise Exception("Cannot unlock, data does not exist:", id)
        # print("LOGGING:::      ...deleted?", id, bool(r))
        return r

    def unlock(self, id, check_success=True):
        """Return whether it succeeded."""
        # print("LOGGING:::  Unlocking...", id)
        r = self.do(self._unlock_, {"id": id}, wait=True)
        if check_success and not r:
            raise Exception("Cannot unlock, data does not exist:", id)
        # print("LOGGING:::      ...unlocked?", id, bool(r))
        return r

    def putdata(self, id, step, inn, stream, parent, locked, ignoredup=False):
        """Return whether it succeeded."""
        # print("LOGGING:::  Putting data...", id)
        r = self.do(self._putdata_, locals(), wait=True)
        # print("LOGGING:::      ...putdata?", id, bool(r))
        return r

    def putstream(self, rows, ignoredup=False):
        """Return whether it succeeded."""
        # print("LOGGING:::  Putting stream...", id)
        r = self.do(self._putstream_, locals(), wait=True)
        # print("LOGGING:::      ...putstream?", id, bool(r))
        return r

    def putcontent(self, id, value, ignoredup=False):
        """Return whether it succeeded."""
        # print("LOGGING:::  Putting content...", id)
        r = self.do(self._putcontent_, locals(), wait=True)
        # print("LOGGING:::      ...putcontent?", id, bool(r))
        return r

    def putfields(self, rows, ignoredup=False):
        """Return number of fields put."""
        # print("LOGGING:::  Putting fields...", rows)
        r = self.do(self._putfields_, locals(), wait=True)
        # print("LOGGING:::      ...put fields?", r, rows)
        return r

    def storestep(self, step, dump=None, ignoredup=False):
        """Return whether it succeeded."""
        return self.putstep(step.id, step.name, step.context, step.config_json, dump and step.dump, ignoredup)

    def putstep(self, id, name, path, config, dump=None, ignoredup=False):
        """Return whether it succeeded."""
        # print("LOGGING:::  Putting step...", id)
        r = self.do(self._putstep_, locals(), wait=True)
        # print("LOGGING:::      ...put step?", id, bool(r))
        return r

    @abstractmethod
    def _hasdata_(self, id, include_empty):
        pass

    @abstractmethod
    def _hasstream_(self, data):
        pass

    @abstractmethod
    def _getdata_(self, id, include_empty):
        pass

    @abstractmethod
    def _getstream_(self, data):
        pass

    @abstractmethod
    def _hasstep_(self, id):
        pass

    @abstractmethod
    def _getstep_(self, id):
        pass

    @abstractmethod
    def _getfields_(self, id):
        pass

    @abstractmethod
    def _hascontent_(self, ids):
        pass

    @abstractmethod
    def _getcontent_(self, id):
        pass

    @abstractmethod
    def _lock_(self, id):
        pass

    @abstractmethod
    def _unlock_(self, id):
        pass

    @abstractmethod
    def _putdata_(self, id, step, inn, stream, parent, locked, ignoredup):
        pass

    @abstractmethod
    def _putstream_(self, rows, ignoredup):
        pass

    @abstractmethod
    def _putfields_(self, rows, ignoredup):
        pass

    @abstractmethod
    def _putcontent_(self, id, value, ignoredup):
        pass

    @abstractmethod
    def _putstep_(self, id, name, path, config, dump, ignoredup):
        pass

    @abstractmethod
    def _deldata_(self, id):
        pass



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import socket

import pymysql
import pymysql.cursors
from pymysql.constants import CLIENT
from sqlalchemy.engine import Engine

from garoupa.decorator import classproperty
from garoupa.uuid import UUID
from tatu.abs.sql import SQL


class SQLA(SQL):
    def _config_(self):
        return self._config

    def __init__(self, engine, threaded=True, close_when_idle=False, storage_info=None, debug=False, read_only=False):
        self._config = locals().copy()
        if not isinstance(engine, Engine):
            raise Exception("Wrong engine:", type(engine))
        del self._config["self"]
        del self._config["__class__"]
        self._uuid = UUID((self.__class__.__name__ + db).encode())
        self.read_only = read_only
        self.engine = engine  # TODO sensitive information should disappear after init
        self.storage_info = storage_info
        self.debug = debug
        self.hostname = socket.gethostname()
        super().__init__(threaded, timeout=8, close_when_idle=close_when_idle)

    def _uuid_(self):
        return self._uuid

    def _open_(self):
        """
        Each reconnection has a cost of approximately 150ms in ADSL (ping=30ms).
        :return:
        """
        if self.debug:
            print("getting connection...")
        self.connection = pymysql.connect(
            host=self.host,
            user=self.user,
            password=self.password,
            charset="utf8",
            # cursorclass=pymysql.cursors.DictCursor,  # set at the bottom of this file
            # client_flag=CLIENT.MULTI_STATEMENTS
        )
        self.connection.client_flag &= pymysql.constants.CLIENT.MULTI_STATEMENTS
        self.connection.autocommit(False)
        # self.connection.server_status

        if self.debug:
            print("getting cursor...")

        # Create db if it doesn't exist yet.
        with self.cursor() as c:
            c.execute(*self.prepare(f"SHOW DATABASES LIKE '{self.db}'"))
            setup = c.fetchone() is None
            if setup:
                if self.debug:
                    print("creating database", self.db, "on", self.database, "...")
                c.execute("create database if not exists " + self.db)
                self.commit()

        if self.debug:
            print("using database", self.db, "on", self.database, "...")
        with self.cursor() as c:
            c.execute("use " + self.db)
            c.execute(f"show tables")

        # Create tables if they don't exist yet.
        try:
            with self.cursor() as c:
                c.execute(f"select 1 from data")
        except:
            if self.debug:
                print("creating database", self.database, "...")
            self._setup()
            self.commit()

        return self

    @classproperty
    def _now_function(cls):
        return "now()"

    @classproperty
    def _keylimit(cls):
        return "(190)"

    @classproperty
    def _auto_incr(cls):
        return "AUTO_INCREMENT"

    @classmethod
    def _on_conflict(cls, cols):
        return "ON DUPLICATE KEY UPDATE"

    @classproperty
    def _insert_ignore(cls):
        return "insert ignore"

    @classmethod
    def _fkcheck(cls, enable):
        return f"SET FOREIGN_KEY_CHECKS={'1' if enable else '0'};"

    @classproperty
    def _placeholder(cls):
        return "%s"

    def newcursor(self):
        return self.connection.cursor(pymysql.cursors.DictCursor)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license. Removing authorship by any means
#  (by code make up or closing the sources) or ignoring property rights
#  is a crime and is unethical regarding the effort and time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
# 
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
# 
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
# 

import threading
from abc import ABC, abstractmethod
from multiprocessing import JoinableQueue, Queue
from queue import Empty


class asThread(ABC):
    # process_lock = None
    thread_lock = None
    queue = None
    outqueue = None
    mythread = None
    isopen = False

    def __init__(self, threaded, timeout, close_when_idle):
        """timeout: Time spent hoping the thread will be useful again."""
        self.isthreaded = threaded
        self.timeout = timeout
        self.close_when_idle = close_when_idle

    def open(self):
        if self.isopen:
            raise Exception("Already open!")
        if self.isthreaded:
            _ = self.threaded  # force start
            return
        self._open_()
        self.isopen = True

    def close(self):
        if self.isthreaded and self.queue:
            self.queue.put(None)
        self.isopen = False
        self.mythread = None

    @property
    def threaded(self):
        if self.isthreaded:
            if self.mythread is None:
                # self.process_lock = Nonemultiprocessing.Lock()
                self.thread_lock = threading.Lock()
                self.queue = Queue()
                self.outqueue = JoinableQueue()
                # self.__class__.mythread = multiprocessing.Process(target=self._worker, daemon=False)
                self.mythread = threading.Thread(target=self._worker, daemon=False)
                #print("LOGGING:::  Starting thread for", self.__class__.__name__)
                self.mythread.start()
        elif not self.isopen:
            self.open()
        return self.isthreaded

    def do(self, func, info, wait=False):
        # Remove self if it came inside 'info' from the caller using locals().
        if "self" in info:
            info = info.copy()
            del info["self"]

        if self.threaded:
            self.queue.put({"func": func.__name__, "info": info, "wait": wait})
            if wait:
                return self._waitresult()
        else:
            return func(**info)

    def _waitresult(self):
        """Wait for result from thread output queue."""
        ret = self.outqueue.get()
        if isinstance(ret, Exception):
            self.outqueue.task_done()
            raise ret
        self.outqueue.task_done()
        return ret

    def _worker(self):
        # TODO: discover why flask reads old values from MySQL until restart,
        #    so we can remove this reconnection from inside the while.

        # conditional 'IF' needed here due to bug #-123
        # File "/home/davi/git/oka-repository/backend/app/api/posts.py", line 319, in get
        #     for m in data.Yt[0]:
        # AttributeError: 'NoneType' object has no attribute 'Yt'
        from tatu.sql.sqlite import SQLite
        if isinstance(self, SQLite):

            try:
                self._open_()
            except Exception as e:
                print(e)
                self.outqueue.put(e)
                raise

        self.isopen = True
        while self.isopen:

            # IF needed due to bug #-123
            # File "/home/davi/git/oka-repository/backend/app/api/posts.py", line 319, in get
            #     for m in data.Yt[0]:
            # AttributeError: 'NoneType' object has no attribute 'Yt'
            if not isinstance(self, SQLite):
                try:
                    self._open_()
                    self.isopen = True
                except Exception as e:
                    print(e)
                    self.outqueue.put(e)
                    raise

            try:
                if self.close_when_idle:
                    # Smart timeout control,so we don't have to wait too much the thread after the main program is gone.
                    t, dt = 0, 0.25
                    job = None
                    while job is None and t < self.timeout:
                        try:
                            job = self.queue.get(timeout=dt)
                        except Empty:
                            if not threading.main_thread().is_alive():
                                break
                        t += dt
                else:
                    job = self.queue.get()
                if job is None:
                    break

                # Handle job from the input queue...
                try:
                    ret = getattr(self, job['func'])(**job["info"])
                    if job["wait"]:
                        self.outqueue.put(ret)
                        self.outqueue.join()
                    # elif "update_remote" in job:
                    #     import dill
                    #     # TODO: Destination Storage needs to be opened in this thread due to SQLite pythonic issues.
                    #     storage = dill.loads(job["update_remote"])
                    #     if storage.isopen:
                    #     raise Exception("Cannot update an already open remote storage on a threaded local storage.")
                    #     self._update_remote_(storage)

                    else:
                        print("Unexpected job:", job)
                except Exception as e:
                    # print(f"Problem while processing job {job}:", e)
                    if threading.main_thread().is_alive():
                        self.outqueue.put(e)
                    break
            except Empty:
                break

            # needed due to bug #-123 explained above
            finally:
                if not isinstance(self, SQLite):
                    self._close_()

        self._close_()

    @abstractmethod
    def _open_(self):
        pass

    @abstractmethod
    def _close_(self):
        pass



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
from abc import ABC
from sqlite3 import IntegrityError as sqliteIntegError

from pymysql import IntegrityError as myIntegError

from akangatu.transf.noop import NoOp
from garoupa.uuid import UUID
from tatu.abs.sqlreadonly import SQLReadOnly
from tatu.abs.storage import LockedEntryException, DuplicateEntryException


class SQL(SQLReadOnly, ABC):
    read_only = False

    def _close_(self):
        from tatu.sql.sqlite import SQLite
        if self.connection and not isinstance(self, SQLite) and self.connection.open:
            self.connection.close()

    def _deldata_(self, id):
        with self.cursor() as c:
            self.run(c, f"delete from data where id=?", [id])
            self.commit()
            r = c.rowcount
        return r == 1

    def _handle_integrity_error(self, id, sql, args):
        try:
            with self.cursor() as c:
                self.run(c, sql, args)
                self.commit()
                r = c.rowcount
            return r == 1
        except (myIntegError, sqliteIntegError) as e:
            with self.cursor() as c2:
                self.run(c2, "select 1 as r from data where id=? and locked=1", [id])
                r2 = c2.fetchone()
                if r2 and "r" in r2:
                    raise LockedEntryException(id, str(e))
                else:
                    raise DuplicateEntryException(id, str(e))

    def _lock_(self, id, ignoredup=False):
        # Placeholder values: step=identity and parent=own-id
        try:
            sql = f"insert {'or ignore' if ignoredup else ''} into data values (null,?,'{NoOp().id}',null,0,?,1)"
            return self._handle_integrity_error(id, sql, [id, id])
        except:
            sql = f"update data set locked=1 where id=?"
            return self._handle_integrity_error(id, sql, [id])

    def _unlock_(self, id):
        with self.cursor() as c:
            self.run(c, self._fkcheck(False))
            try:
                self.run(c, "delete from data where id=? and locked=1", [id])
                r = c.rowcount
            finally:
                self.run(c, self._fkcheck(True))
                self.commit()
            return r == 1

    def _putdata_(self, id, step, inn, stream, parent, locked, ignoredup=False):
        # if id == "0iQlsaDROpjL7YRZBcFr0Py":
        #     raise Exception("--------------------")
        sql = f"insert {'or ignore' if ignoredup else ''} INTO data values (null,?,?,?,?,?,?)"
        return self._handle_integrity_error(id, sql, [id, step, inn, stream, parent, locked])

    def _putstream_(self, rows, ignoredup=False):
        # try:
        with self.cursor() as c:
            self.write_many(c, rows, "stream", ignoredup)
            self.commit()
            r = c.rowcount
        return r

    # except Exception as e:
    #     import traceback
    #     traceback.print_exc()
    #     print(e)

    def _putfields_(self, rows, ignoredup=False):
        with self.cursor() as c:
            self.write_many(c, rows, "field", ignoredup)
            self.commit()
            return c.rowcount

    def _putcontent_(self, id, value, ignoredup=False):
        with self.cursor() as c:
            self.run(c, f"insert {'or ignore' if ignoredup else ''} INTO content VALUES (?,?)", [id, value])
            self.commit()
            return c.rowcount == 1

    def _putstep_(self, id, name, path, config, dump=None, ignoredup=False):
        configid = UUID(config.encode()).id
        # ALmost never two steps will have the same config,
        #   except the shortest ones which render worthless the avoidance of a second 'insert' attempt.
        with self.cursor() as c:
            self.run(c, f"insert or ignore INTO config VALUES (?,?)", [configid, config])
            sql = f"insert {'or ignore' if ignoredup else ''} INTO step VALUES (NULL,?,?,?,?,?)"
            self.run(c, sql, [id, name, path, configid, dump])
            self.commit()
            return c.rowcount == 1

    def write_many(self, cursor, list_of_tuples, table, ignore_dup=True):
        command = self._insert_ignore if ignore_dup else 'insert'
        sql = f"{command} INTO {table} VALUES({('?,' * len(list_of_tuples[0]))[:-1]})"

        newlist_of_tuples = []
        for row in list_of_tuples:
            newrow = [int(c) if isinstance(c, bool) else c for c in row]
            newlist_of_tuples.append(newrow)
            if self.debug:
                msg = self._interpolate(sql, newrow)
                print(self.name + ":\t>>>>> " + msg)

        sql = sql.replace("?", self._placeholder)
        return cursor.executemany(sql, newlist_of_tuples)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
from abc import ABC

from tatu.abs.mixin.setup import withSetup
from tatu.abs.storage import LockedEntryException
from tatu.storageinterface import StorageInterface


class SQLReadOnly(StorageInterface, withSetup, ABC):
    read_only = True

    def _hasdata_(self, id, include_empty):  # TODO checar lock null?
        if include_empty:
            sql, args = f"select 1 from data where id=?", [id]
        else:
            # REMINDER Inner join ensures a Data row with fields.
            nonempty = f"select 1 from data d INNER JOIN field f ON d.id=f.data where d.id=? limit 1"
            withstream = "select 1 from data where id=? and stream=1"
            sql, args = f"{withstream} UNION {nonempty}", [id, id]
        with self.cursor() as c:
            self.run(c, sql, args)
            return c.fetchone() is not None

    def _hasstream_(self, data):
        sql = f"select 1 from stream where id=? and pos=0"
        with self.cursor() as c:
            self.run(c, sql, [data])
            return c.fetchone() is not None

    def _getdata_(self, id, include_empty):
        cols = "step,inn,stream,parent,locked,name as field_name,content as field_id"
        sql = f"select {cols} from data d {'left' if include_empty else 'inner'} join field f on d.id=f.data where d.id=?"
        with self.cursor() as c:
            self.run(c, sql, [id])
            r = c.fetchall()
        if not r:
            return
        uuids = {}
        for row in r:
            if row["locked"]:
                raise LockedEntryException("Cannot get a locked Data object.")
            if row["field_name"]:
                uuids[row["field_name"]] = row["field_id"]
        return {"uuids": uuids, "step": row["step"], "parent": row["parent"], "inner": row["inn"],
                "stream": row["stream"]}

    def _getstream_(self, data):
        sql = "select data,pos,chunk from stream where data=?"
        with self.cursor() as c:
            self.run(c, sql, [data])
            # TODO: can this be passed around as an iterator? is it worth or will just increase the amount of latencies?
            row = c.fetchall()
        if len(row) == 0:
            return
        return row

    def _getstep_(self, id):
        sql = "select name,path,params from step s inner join config c on s.config=c.id where s.id=?"
        with self.cursor() as c:
            self.run(c, sql, [id])
            row = c.fetchone()
        if row is None:
            return
        desc = {"name": row["name"], "path": row["path"], "config": row["params"]}
        return desc

    def _getfields_(self, id):
        sql = f"select content as cid,value from field inner join content on content=id where data=?"
        with self.cursor() as c:
            self.run(c, sql, id)
            rows = c.fetchall()
        return {row["cid"]: row["value"] for row in
                rows}  # TODO retornar iterator; pra isso, precisa de uma conexão fora da thread, e gets são bloqueantes anyway

    def _getcontent_(self, id):
        sql = f"select value from content where id=?"
        with self.cursor() as c:
            self.run(c, sql, [id])
            row = c.fetchone()
        return row and row["value"]

    def _hasstep_(self, id):
        sql = f"select 1 from step where id=?"
        with self.cursor() as c:
            self.run(c, sql, [id])
            return c.fetchone() is not None

    def _hascontent_(self, ids):
        sql = f"select id from content where id in ({('?,' * len(ids))[:-1]})"  # REMINDER: -1 deletes comma
        with self.cursor() as c:
            self.run(c, sql, ids)
            return [r["id"] for r in c.fetchall()]

    # #     #     # From a StackOverflow answer...
    # #     #     import sys
    # #     #     import traceback
    # #     #
    # #     #     msg = "STORAGE DBG:" + self.info + "\n" + msg
    # #     #     # Gather the information from the original exception:
    # #     #     exc_type, exc_value, exc_traceback = sys.exc_info()
    # #     #     # Format the original exception for a nice printout:
    # #     #     traceback_string = "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
    # #     #     # Re-raise a new exception of the same class as the original one
    # #     #     raise type(ex)("%s\norig. trac.:\n%s\n" % (msg, traceback_string))

    # noinspection PyDefaultArgument
    def prepare(self, sql, args=[]):
        # TODO with / catch / finalize connection?
        args = [int(c) if isinstance(c, bool) else c for c in args]
        sql = sql.replace("insert or ignore", self._insert_ignore)
        if self.debug:
            msg = self._interpolate(sql, args)
            print(self.name + ":\t>>>>> " + msg)
        # self.connection.ping(reconnect=True)
        return sql.replace("?", self._placeholder), args

    @staticmethod
    def _interpolate(sql, lst0):
        lst = [str(w)[:100] for w in lst0]
        zipped = zip(sql.replace("?", '"?"').split("?"), map(str, lst + [""]))
        return "".join(list(sum(zipped, ()))).replace('"None"', "NULL")

    def commit(self):
        #print("LOGGING:::  commit")
        self.connection.commit()



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
from abc import abstractmethod, ABC
from contextlib import contextmanager

from aiuna.content.root import Root
from akangatu.transf.noop import NoOp
from garoupa.decorator import classproperty
from garoupa.uuid import UUID


class Setup:
    """
    >>> 1
    1
    """


class withSetup(ABC):
    connection = None

    @contextmanager
    def cursor(self):
        cursor = self.newcursor()
        try:
            yield cursor
        finally:
            cursor.close()

    def run(self, cursor, sql, args=[]):
        return cursor.execute(*self.prepare(sql, args))

    @abstractmethod
    def newcursor(self):
        pass

    @abstractmethod
    def prepare(self, sql, arg=[]):
        pass

    @classmethod
    @abstractmethod
    @classproperty
    def _now_function(cls):
        pass

    @classmethod
    @abstractmethod
    @classproperty
    def _keylimit(cls):
        pass

    @classmethod
    @abstractmethod
    @classproperty
    def _auto_incr(cls):
        pass

    @classmethod
    @abstractmethod
    def _fkcheck(cls, enable):
        pass

    @classmethod
    @abstractmethod
    def _on_conflict(cls, cols):
        pass

    @classmethod
    @classproperty
    @abstractmethod
    def _insert_ignore(cls):
        pass

    @classmethod
    @abstractmethod
    @classproperty
    def _placeholder(cls):
        pass

    def _setup(self):
        print("creating tables...")
        # REMINDER 'inner' is a reserved SQL keyword.
        # REMINDER d.parent = d.uuid / step.uuid; the field is there just to allow a single recursive query directly on
        # the server.
        # REMINDER char automatically pads with spaces but ignores them when comparing, perfect instead of varchar that
        # is slower
        # REMINDER field char(24) is almost arbitrary, perhaps someone wants to name a field with a uuid;
        #      23+1 just shows that it is not a uuid column.

        # Table for field contents (matrices, ...) and dumps (pickled models, ...).
        sql = "create table if not exists content (id char(23) NOT NULL primary key, value LONGBLOB NOT NULL)"
        with self.cursor() as c:
            self.run(c, sql)

        # Table with values for parameters of Step objects. Insert the empty config as the fisrt one.
        sql = "create table if not exists config (id char(23) NOT NULL primary key, params text NOT NULL)"
        with self.cursor() as c:
            self.run(c, sql)
            try:  # REMINDER Only MariaDB accepts 'CREATE INDEX if not exists'
                self.run(c, f"CREATE INDEX  idx1 ON config (params{self._keylimit})")
            except:
                pass
            self.run(c, f"insert into config values ('{UUID(b'{}').id}', " + "'{}')")

        # Table with steps.
        # The mythical NoOp step that created the Root data is inserted as the first one (due to constraint issues).
        sql = f"""
            create table if not exists step (
                n integer NOT NULL primary key {self._auto_incr},
                id char(23) NOT NULL UNIQUE,
                name char(60) NOT NULL,
                path varchar(250) NOT NULL,
                config char(23) NOT NULL,
                content char(23),
                FOREIGN KEY (config) REFERENCES config(id),
                FOREIGN KEY (content) REFERENCES content(id)
            )"""
        with self.cursor() as c:
            self.run(c, sql)
            try:
                self.run(c, f'CREATE INDEX idx2 ON step (id)')
                self.run(c, f'CREATE INDEX idx3 ON step (name)')
                self.run(c, f'CREATE INDEX idx4 ON step (path)')
            except:
                pass
        no = NoOp()
        sql = f"insert into step values (null, '{no.id}', '{no.name}', '{no.context}', '{UUID(b'{}').id}', null)"
        with self.cursor() as c:
            self.run(c, sql)

        # Table data.
        sql = f"""
            create table if not exists data (
                n integer NOT NULL primary key {self._auto_incr},
                id char(23) NOT NULL UNIQUE,
                step char(23) NOT NULL,
                inn char(23),
                stream boolean not null,
                parent char(23) not null,
                locked boolean,
                unique(step, parent),
                FOREIGN KEY (step) REFERENCES step(id),
                FOREIGN KEY (inn) REFERENCES data(id),
                FOREIGN KEY (parent) REFERENCES data(id)
            )"""
        with self.cursor() as c:
            self.run(c, sql)
            try:
                self.run(c, f'CREATE INDEX  idx5 ON data (id)')
                self.run(c, f'CREATE INDEX  idx6 ON data (step)')
                self.run(c, f'CREATE INDEX  idx7 ON data (parent)')
            except:
                pass
        sql = f"insert into data values (null, '{Root.id}', '{UUID.identity.id}', null, 0, '{Root.id}', 0)"
        with self.cursor() as c:
            self.run(c, sql)

        sql = f"""
            create table if not exists field (
                data char(23) NOT NULL,
                name char(24) NOT NULL,
                content char(23) NOT NULL,
                primary key (data, name),
                FOREIGN KEY (data) REFERENCES data(id),
                FOREIGN KEY (content) REFERENCES content(id)
            )"""
        with self.cursor() as c:
            self.run(c, sql)
            try:
                self.run(c, f'CREATE INDEX  idx8 ON field (name)')
                self.run(c, f'CREATE INDEX  idx9 ON field (data)')
            except:
                pass

        # Table to speed up lookup for not yet synced Data objects.
        sql = f"""
            create table if not exists storage (
                id char(23) NOT NULL primary key,
                data char(23) NOT NULL,
                t DATETIME,
                FOREIGN KEY (data) REFERENCES data(id)
            )"""
        with self.cursor() as c:
            self.run(c, sql)
            try:
                self.run(c, f'CREATE INDEX  idx10 ON storage (data)')
            except:
                pass

        # Table to record stream, like folds of cross-validation (or chunks of a datastream?).
        sql = f"""
            create table if not exists stream (
                data char(23) NOT NULL,
                pos integer not null,
                chunk char(23) NOT NULL,
                unique (data, pos),
                FOREIGN KEY (data) REFERENCES data(id),
                FOREIGN KEY (chunk) REFERENCES data(id)
            )"""
        with self.cursor() as c:
            self.run(c, sql)
            try:
                self.run(c, f'CREATE INDEX  idx11 ON stream (chunk)')
            except:
                pass

        # Table to record volatile info, i.e. related to a specific run of a step*
        #     (e.g. person specific or several runs to assess time).
        # 'id' here is a universal time based UUID(),
        #   instead of being based on a hash or on a multiplication like the other ones.
        # * -> related to a specific generation of a Data object - more precisely.
        #  Many rows can exist for the same Data object.
        # 'alive': time of last ping from 'node'
        # 'duration'=NULL while not finished
        sql = f"""
            create table if not exists run (
                id char(23) NOT NULL primary key,
                data char(23) NOT NULL,
                duration float,
                node char(255) NOT NULL,
                alive DATETIME,
                t DATETIME,
                FOREIGN KEY (data) REFERENCES data(id)
            )"""
        with self.cursor() as c:
            self.run(c, sql)
            try:
                self.run(c, f'CREATE INDEX  idx12 ON run (data)')
                self.run(c, f'CREATE INDEX  idx13 ON run (node)')
            except:
                pass

        # fail TINYINT
        # update data set {','.join([f'{k}=?' for k in to_update.keys()])}
        #     x = coalesce(values(x), x),   # Return the first non-null value in a list
        #     from res left join data on dout = did
        #     left join dataset on dataset = dsid where



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license. More about this in the section (*) below.
#  Relevant employers or funding agencies will be notified accordingly.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#

import setuptools

NAME = "tatu"


VERSION = "0.2102.11"


AUTHOR = 'Davi Pereira dos Santos'


AUTHOR_EMAIL = 'dpsbac@gmail.com'


DESCRIPTION = 'Persistent Science'


with open('README.md', 'r') as fh:
    LONG_DESCRIPTION = fh.read()


LICENSE = 'GPL3'


URL = 'https://github.com/davips/tatu'


DOWNLOAD_URL = 'https://github.com/davips/tatu/releases'


CLASSIFIERS = ['Intended Audience :: Science/Research',
               'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
               'Natural Language :: English',
               'Programming Language :: Python',
               'Topic :: Scientific/Engineering',
#posix               'Operating System :: Linux',
               'Programming Language :: Python :: 3.8']


INSTALL_REQUIRES = [
    'numpy', 'sklearn', 'liac-arff', 'pymysql', 'aiuna', 'requests',
    'sqlalchemy'
]


EXTRAS_REQUIRE = {
}

SETUP_REQUIRES = ['wheel']

setuptools.setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    classifiers=CLASSIFIERS,
    description=DESCRIPTION,
    download_url=DOWNLOAD_URL,
    extras_require=EXTRAS_REQUIRE,
    install_requires=INSTALL_REQUIRES,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license=LICENSE,
    packages=setuptools.find_packages(),
    setup_requires=SETUP_REQUIRES,
    url=URL,
)

package_dir = {'': 'tatu'}  # For IDEs like Intellij to recognize the package.




#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
from abc import ABC, abstractmethod

from aiuna.content.data import Data
from akangatu.transf.mixin.identification import withIdentification


class Storage(withIdentification, ABC):
    """
    Store and recover results from some place.
    The children classes are expected to provide storage in e.g.:
     SQLite, remote/local MongoDB, MySQL server, pickled or even CSV files.
    """

    @property
    def config(self):
        return self._config_()

    @abstractmethod
    def _config_(self):
        pass

    @property
    def desc(self):
        return {"name": self.name, "path": self.context, "config": self.config}

    @property
    def asdict(self):
        return {"id": self.id, "desc": self.desc}

    @abstractmethod
    def fetch(self, data, lock=False, lazy=True):  # , recursive=True):
        """Fetch the Data object fields on-demand if lazy=True.
         data: uuid string or a (probably still not fully evaluated) Data object."""

    @abstractmethod
    def store(self, data: Data, unlock=False, ignoredup=False, lazy=False):
        """Store all Data object fields as soon as one of them is evaluated.

        # The sequence of queries is planned to minimize traffic and CPU load,
        # otherwise it would suffice to just send 'insert or ignore' of dumps.

        Parameters
        ----------
        data
            Data object to store.
        ignoredup
            Whether to send the query anyway, ignoring errors due to already existent registries.

        Returns
        -------
        List of inserted (or hoped to be inserted for threaded storages) Data ids (only meaningful for Data objects with inner)

        Exception
        ---------
        DuplicateEntryException
        :param ignoredup:
        :param data:
        :param check_dup:
        :param recursive:
        """

    @abstractmethod
    def fetchhistory(self, id):
        """Fetch History for the given Data object id."""

    @abstractmethod
    def fetchstep(self, id):
        """Return a Step object."""

    @abstractmethod
    def hasdata(self, id, include_empty=False):
        """ include_empty: whether to assess the existence of fields, instead of just the data row"""

    @abstractmethod
    def hasstream(self, data):
        """Verify existence of stream for a given dataid."""

    @abstractmethod
    def getdata(self, id, include_empty=True):
        """Return info for a Data object."""

    @abstractmethod
    def getstream(self, data):
        """Return rows with info for the stream of a given Data id."""

    @abstractmethod
    def hasstep(self, id):
        """Does the step with 'id' exist?"""

    @abstractmethod
    def getstep(self, id):
        """Return info for a Step object."""

    @abstractmethod
    def getfields(self, id):
        """Return fields and content for a Data object."""

    @abstractmethod
    def getcontent(self, id):
        """Return content with the given id."""

    @abstractmethod
    def hascontent(self, ids):
        """Does the content with 'id' exist?"""

    @abstractmethod
    def removedata(self, data: Data, check_existence=True, recursive=True):
        """Remove Data object, but keeps contents of its fields (even if not used by anyone else).

        Returns list of deleted Data object uuids
        """

    @abstractmethod
    def lock(self, id, check_existence=True):
        """Return whether it succeeded."""

    @abstractmethod
    def deldata(self, id, check_success=True):
        """Return whether it succeeded."""

    @abstractmethod
    def unlock(self, id, check_success=True):
        """Return whether it succeeded."""

    @abstractmethod
    def putstream(self, rows, ignoredup=False):
        """Return whether it succeeded."""

    @abstractmethod
    def putdata(self, id, step, inn, stream, parent, locked, ignoredup=False):
        """Return whether it succeeded."""

    @abstractmethod
    def putcontent(self, id, value, ignoredup=False):
        """Return whether it succeeded."""

    @abstractmethod
    def putfields(self, rows, ignoredup=False):
        """Return whether it succeeded."""

    @abstractmethod
    def storestep(self, step, dump=None, ignoredup=False):
        """Return whether it succeeded."""

    @abstractmethod
    def putstep(self, id, name, path, config, dump=None, ignoredup=False):
        """Insert a step registry with the given fields"""

    def _name_(self):
        return self.__class__.__name__

    def _context_(self):
        return self.__class__.__module__


class UnlockedEntryException(Exception):
    """No locked entry for this input data."""


class LockedEntryException(Exception):
    """Another node is/was generating output data for this input data."""


class DuplicateEntryException(Exception):
    """This input data has already been inserted before."""


class MissingEntryException(Exception):
    """This input data is missing."""
